package logger_test

import (
	"bytes"
	"fmt"
	"net/http"
	"strings"
	"sync"
	"testing"

	"codeberg.org/gruf/go-kv"
	"codeberg.org/gruf/go-logger/v2/level"
	"codeberg.org/gruf/go-logger/v2/log"
)

var (
	testArgs = [][]interface{}{
		{0, 1, 2, 3, 4},
		{"hello", "world!"},
		{"random", 0, 1, 2.2, 3.3, "arguments"},
		{[]int{0, 1, 2, 3}, []string{"hello", "world"}, map[string]string{"hello": "world"}},
		{http.Client{}, sync.Mutex{}},
		{nil, nil, nil, nil},
	}

	testFmts = []struct {
		fmt string
		arg []interface{}
	}{
		{
			fmt: "hello world: %q",
			arg: []interface{}{"hello world"},
		},
		{
			fmt: "no operators here",
			arg: []interface{}{"oh no unexpected!"},
		},
		{
			fmt: "0, 1, %d, %f, %s",
			arg: []interface{}{2, 3.0, "4"},
		},
		{
			fmt: "%d %d %d %d",
			arg: []interface{}{"oh no bad and missing operators"},
		},
		{
			fmt: "%#v",
			arg: []interface{}{http.Client{}},
		},
		{
			fmt: "%+v",
			arg: []interface{}{map[string]interface{}{"key": "value", "hello": "world"}},
		},
	}

	testFields = []kv.Fields{
		{
			kv.Field{K: "hello", V: "world!"},
			kv.Field{K: "client", V: http.Client{}},
		},
		{
			kv.Field{K: "map", V: map[string]string{"map": "of values"}},
			kv.Field{K: "slice", V: []string{"slice", "of", "values"}},
		},
		{
			kv.Field{K: "nil", V: nil},
			kv.Field{K: "hmmm", V: "watcha sayyyy"},
		},
		{
			kv.Field{K: "int", V: 420},
			kv.Field{K: "float", V: 6.9},
		},
		{
			kv.Field{K: "complex", V: complex(120, 120)},
			kv.Field{K: "shrug", V: "\n\n\n\n\n\n\n\n\n\n\n\n"},
		},
	}
)

func TestLimitByLevel(t *testing.T) {
	// Special case of no log flags
	log.SetFlags(0)

	// Output to buffer
	buf := bytes.Buffer{}
	log.SetOutput(&buf)

	testLog := func(do func(), expect bool) {
		do()
		if !expect && buf.Len() > 0 {
			t.Error("log should not have produced output")
		} else if expect && buf.Len() == 0 {
			t.Error("log should have produced output")
		}
		buf.Reset()
	}

	for _, lvl := range []level.LEVEL{
		level.ALL,
		level.TRACE,
		level.DEBUG,
		level.INFO,
		level.WARN,
		level.ERROR,
		level.FATAL,
		level.PANIC,
		level.UNSET,
	} {
		// Update log level
		log.SetLevel(lvl)

		// TRACE level
		testLog(func() {
			log.Trace("test log")
		}, lvl.CanLog(level.TRACE))
		testLog(func() {
			log.Tracef("test log")
		}, lvl.CanLog(level.TRACE))
		testLog(func() {
			log.TraceKVs(kv.Field{K: "test", V: "log"})
		}, lvl.CanLog(level.TRACE))

		// DEBUG level
		testLog(func() {
			log.Debug("test log")
		}, lvl.CanLog(level.DEBUG))
		testLog(func() {
			log.Debugf("test log")
		}, lvl.CanLog(level.DEBUG))
		testLog(func() {
			log.DebugKVs(kv.Field{K: "test", V: "log"})
		}, lvl.CanLog(level.DEBUG))

		// INFO level
		testLog(func() {
			log.Info("test log")
		}, lvl.CanLog(level.INFO))
		testLog(func() {
			log.Infof("test log")
		}, lvl.CanLog(level.INFO))
		testLog(func() {
			log.InfoKVs(kv.Field{K: "test", V: "log"})
		}, lvl.CanLog(level.INFO))

		// WARN level
		testLog(func() {
			log.Warn("test log")
		}, lvl.CanLog(level.WARN))
		testLog(func() {
			log.Warnf("test log")
		}, lvl.CanLog(level.WARN))
		testLog(func() {
			log.WarnKVs(kv.Field{K: "test", V: "log"})
		}, lvl.CanLog(level.WARN))

		// ERROR level
		testLog(func() {
			log.Error("test log")
		}, lvl.CanLog(level.ERROR))
		testLog(func() {
			log.Errorf("test log")
		}, lvl.CanLog(level.ERROR))
		testLog(func() {
			log.ErrorKVs(kv.Field{K: "test", V: "log"})
		}, lvl.CanLog(level.ERROR))

		// FATAL level
		// can't do because this will syscall exit :)

		// PANIC level
		testLog(func() {
			defer func() { _ = recover() }()
			log.Panic("test log")
		}, lvl.CanLog(level.PANIC))
		testLog(func() {
			defer func() { _ = recover() }()
			log.Panicf("test log")
		}, lvl.CanLog(level.PANIC))
		testLog(func() {
			defer func() { _ = recover() }()
			log.PanicKVs(kv.Field{K: "test", V: "log"})
		}, lvl.CanLog(level.PANIC))
	}
}

func TestLevelOutput(t *testing.T) {
	// Special case of no log flags
	log.SetFlags(0)

	// Output to buffer
	buf := bytes.Buffer{}
	log.SetOutput(&buf)

	// Reset log level
	log.SetLevel(level.ALL)

	testLog := func(do func(), match func(string) bool) {
		do()
		if str := buf.String(); !match(strings.TrimSpace(str)) {
			t.Errorf("unexpected log output: %s", str)
		}
		buf.Reset()
	}

	for _, arg := range testArgs {
		expect := fmt.Sprint(arg...)

		testLog(func() {
			log.Debug(arg...)
		}, func(out string) bool {
			return strings.HasPrefix(out, "[DEBUG] ") &&
				(strings.TrimPrefix(out, "[DEBUG] ") == expect)
		})

		testLog(func() {
			log.Info(arg...)
		}, func(out string) bool {
			return strings.HasPrefix(out, "[INFO] ") &&
				(strings.TrimPrefix(out, "[INFO] ") == expect)
		})

		testLog(func() {
			log.Warn(arg...)
		}, func(out string) bool {
			return strings.HasPrefix(out, "[WARN] ") &&
				(strings.TrimPrefix(out, "[WARN] ") == expect)
		})

		testLog(func() {
			log.Error(arg...)
		}, func(out string) bool {
			return strings.HasPrefix(out, "[ERROR] ") &&
				(strings.TrimPrefix(out, "[ERROR] ") == expect)
		})

		// FATAL level
		// can't do because this will syscall exit :)

		testLog(func() {
			defer func() {
				if recover() != expect {
					t.Error("unexpected panic result")
				}
			}()
			log.Panic(arg...)
		}, func(out string) bool {
			return strings.HasPrefix(out, "[PANIC] ") &&
				(strings.TrimPrefix(out, "[PANIC] ") == expect)
		})
	}

	for _, format := range testFmts {
		expect := fmt.Sprintf(format.fmt, format.arg...)

		testLog(func() {
			log.Debugf(format.fmt, format.arg...)
		}, func(out string) bool {
			return strings.HasPrefix(out, "[DEBUG] ") &&
				(strings.TrimPrefix(out, "[DEBUG] ") == expect)
		})

		testLog(func() {
			log.Infof(format.fmt, format.arg...)
		}, func(out string) bool {
			return strings.HasPrefix(out, "[INFO] ") &&
				(strings.TrimPrefix(out, "[INFO] ") == expect)
		})

		testLog(func() {
			log.Warnf(format.fmt, format.arg...)
		}, func(out string) bool {
			return strings.HasPrefix(out, "[WARN] ") &&
				(strings.TrimPrefix(out, "[WARN] ") == expect)
		})

		testLog(func() {
			log.Errorf(format.fmt, format.arg...)
		}, func(out string) bool {
			return strings.HasPrefix(out, "[ERROR] ") &&
				(strings.TrimPrefix(out, "[ERROR] ") == expect)
		})

		// FATAL level
		// can't do because this will syscall exit :)

		testLog(func() {
			defer func() {
				if recover() != expect {
					t.Error("unexpected panic result")
				}
			}()
			log.Panicf(format.fmt, format.arg...)
		}, func(out string) bool {
			return strings.HasPrefix(out, "[PANIC] ") &&
				(strings.TrimPrefix(out, "[PANIC] ") == expect)
		})
	}

	for _, fields := range testFields {
		expect := fields.String()

		testLog(func() {
			log.DebugKVs(fields...)
		}, func(out string) bool {
			return strings.HasPrefix(out, "[DEBUG] ") &&
				(strings.TrimPrefix(out, "[DEBUG] ") == expect)
		})

		testLog(func() {
			log.InfoKVs(fields...)
		}, func(out string) bool {
			return strings.HasPrefix(out, "[INFO] ") &&
				(strings.TrimPrefix(out, "[INFO] ") == expect)
		})

		testLog(func() {
			log.WarnKVs(fields...)
		}, func(out string) bool {
			return strings.HasPrefix(out, "[WARN] ") &&
				(strings.TrimPrefix(out, "[WARN] ") == expect)
		})

		testLog(func() {
			log.ErrorKVs(fields...)
		}, func(out string) bool {
			return strings.HasPrefix(out, "[ERROR] ") &&
				(strings.TrimPrefix(out, "[ERROR] ") == expect)
		})

		// FATAL level
		// can't do because this will syscall exit :)

		testLog(func() {
			defer func() {
				if recover() != expect {
					t.Error("unexpected panic result")
				}
			}()
			log.PanicKVs(fields...)
		}, func(out string) bool {
			return strings.HasPrefix(out, "[PANIC] ") &&
				(strings.TrimPrefix(out, "[PANIC] ") == expect)
		})
	}
}
